from polars import Expr


def AVGPRICE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """(open + high + low + close) / 4

    References
    ----------
    https://github.com/TA-Lib/ta-lib/blob/main/src/ta_func/ta_AVGPRICE.c#L187

    """
    return (open + high + low + close) / 4


def MEDPRICE(high: Expr, low: Expr) -> Expr:
    """(high + low) / 2

    References
    ----------
    https://github.com/TA-Lib/ta-lib/blob/main/src/ta_func/ta_MEDPRICE.c#L180

    """
    return (high + low) / 2


def TYPPRICE(high: Expr, low: Expr, close: Expr) -> Expr:
    """(high + low + close) / 3

    References
    ----------
    https://github.com/TA-Lib/ta-lib/blob/main/src/ta_func/ta_TYPPRICE.c#L185

    """
    return (high + low + close) / 3


def WCLPRICE(high: Expr, low: Expr, close: Expr) -> Expr:
    """(high + low + close * 2) / 4

    References
    ----------
    https://github.com/TA-Lib/ta-lib/blob/main/src/ta_func/ta_WCLPRICE.c#L184

    """
    return (high + low + close * 2) / 4



import polars_ols as pls
from polars import Expr, Int32, UInt16, struct, when
from polars import arange, repeat
from polars import rolling_corr, rolling_cov
from polars_ols import RollingKwargs

from polars_ta.utils.numba_ import batches_i1_o1, batches_i2_o1, batches_i2_o2
from polars_ta.utils.pandas_ import roll_kurt, roll_rank
from polars_ta.wq._nb import roll_argmax, roll_argmin, roll_prod, roll_co_kurtosis, roll_co_skewness, roll_moment, roll_partial_corr, roll_triple_corr, _cum_prod_by, _cum_sum_by, _signals_to_size, _cum_sum_reset, _sum_split_by


def ts_arg_max(x: Expr, d: int = 5, reverse: bool = True) -> Expr:
    """Returns the relative index of the max value in the time series for the past d days.
    If the current day has the max value for the past d days, it returns 0.
    If previous day has the max value for the past d days, it returns 1.

    Parameters
    ----------
    x
    d
    reverse
        反向

    See Also
    --------
    ts_arg_min

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [6, 2, 8, 5, 9, 4][::-1],
    }).with_columns(
        out=ts_arg_max(pl.col('a'), 6),
    )
    shape: (6, 2)
    ┌─────┬──────┐
    │ a   ┆ out  │
    │ --- ┆ ---  │
    │ i64 ┆ u16  │
    ╞═════╪══════╡
    │ 4   ┆ null │
    │ 9   ┆ null │
    │ 5   ┆ null │
    │ 8   ┆ null │
    │ 2   ┆ null │
    │ 6   ┆ 4    │
    └─────┴──────┘
    ```

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#ts_arg_maxx-d

    """
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), roll_argmax, d, reverse, dtype=UInt16))


def ts_arg_min(x: Expr, d: int = 5, reverse: bool = True) -> Expr:
    """

    Parameters
    ----------
    x
    d
    reverse
        反向

    See Also
    --------
    ts_arg_max

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#ts_arg_minx-d

    """
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), roll_argmin, d, reverse, dtype=UInt16))


def ts_co_kurtosis(x: Expr, y: Expr, d: int = 5, ddof: int = 0) -> Expr:
    return struct([x, y]).map_batches(lambda xx: batches_i2_o1([xx.struct[i].to_numpy() for i in range(2)], roll_co_kurtosis, d))


def ts_co_skewness(x: Expr, y: Expr, d: int = 5, ddof: int = 0) -> Expr:
    return struct([x, y]).map_batches(lambda xx: batches_i2_o1([xx.struct[i].to_numpy() for i in range(2)], roll_co_skewness, d))


def ts_corr(x: Expr, y: Expr, d: int = 5, ddof: int = 1) -> Expr:
    """rolling correlation between two columns

    时序滚动相关系数

    Parameters
    ----------
    x
    y
    d
    ddof
        自由度

    Notes
    -----
    x、y不区分先后

    """
    return rolling_corr(x, y, window_size=d, ddof=ddof)


def ts_count(x: Expr, d: int = 30) -> Expr:
    """时序滚动计数"""
    return x.cast(Int32).rolling_sum(d)


def ts_count_nans(x: Expr, d: int = 5) -> Expr:
    """时序滚动统计nan出现次数"""
    return x.is_nan().rolling_sum(d)


def ts_count_nulls(x: Expr, d: int = 5) -> Expr:
    """时序滚动统计null出现次数"""
    return x.is_null().rolling_sum(d)


def ts_covariance(x: Expr, y: Expr, d: int = 5, ddof: int = 1) -> Expr:
    """rolling covariance between two columns

    时序协方差

    Parameters
    ----------
    x
    y
    d
    ddof
        自由度

    Notes
    -----
    x、y不区分先后

    """
    return rolling_cov(x, y, window_size=d, ddof=ddof)


def ts_cum_max(x: Expr) -> Expr:
    """时序累计最大"""
    return x.cum_max()


def ts_cum_min(x: Expr) -> Expr:
    """时序累计最小"""
    return x.cum_min()


def ts_cum_prod(x: Expr) -> Expr:
    """时序累乘"""
    return x.cum_prod()


def ts_cum_sum(x: Expr) -> Expr:
    """时序累加"""
    return x.cum_sum()


def ts_cum_sum_reset(x: Expr) -> Expr:
    """时序累加。遇到0、nan、相反符号时重置

    Examples
    --------

    ```python
    df = pl.DataFrame({
        'a': [1, 0, 1, 2, None, 3, -2, -3],
    }).with_columns(
        A=ts_cum_sum_reset(pl.col('a'))
    )

    shape: (8, 2)
    ┌──────┬──────┐
    │ a    ┆ A    │
    │ ---  ┆ ---  │
    │ i64  ┆ f64  │
    ╞══════╪══════╡
    │ 1    ┆ 1.0  │
    │ 0    ┆ 0.0  │
    │ 1    ┆ 1.0  │
    │ 2    ┆ 3.0  │
    │ null ┆ 0.0  │
    │ 3    ┆ 3.0  │
    │ -2   ┆ -2.0 │
    │ -3   ┆ -5.0 │
    └──────┴──────┘

    ```

    """
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy().astype(float), _cum_sum_reset))


def ts_decay_exp_window(x: Expr, d: int = 30, factor: float = 1.0) -> Expr:
    """指数衰减移动平均

    Parameters
    ----------
    x
    d
    factor
        衰减系数

    Warnings
    --------
    weights not yet supported on array with null values

    """
    y = arange(d - 1, -1, step=-1, eager=False)
    weights = repeat(factor, d, eager=True).pow(y)
    return x.rolling_mean(d, weights=weights)


def ts_decay_linear(x: Expr, d: int = 30) -> Expr:
    """线性衰减移动平均

    Warnings
    --------
    weights not yet supported on array with null values

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#ts_decay_linearx-d-dense-false

    """
    weights = arange(1, d + 1, eager=True)
    return x.rolling_mean(d, weights=weights)


def ts_delay(x: Expr, d: int = 1, fill_value=None) -> Expr:
    """时序数据移动 shift x

    Parameters
    ----------
    x
    d
        向前或向后的移动天数
    fill_value
        填充。可用None、常量或Expr

    """
    return x.shift(d, fill_value=fill_value)


def ts_delta(x: Expr, d: int = 1) -> Expr:
    """差分"""
    return x.diff(d)


def ts_fill_null(x: Expr) -> Expr:
    """用上一个非空值填充空值"""
    return x.forward_fill()


def ts_ir(x: Expr, d: int = 1) -> Expr:
    """时序滚动信息系数rolling information ratio"""
    return ts_mean(x, d) / ts_std_dev(x, d, 0)


def ts_kurtosis(x: Expr, d: int = 5) -> Expr:
    """kurtosis of x for the last d days

    时序滚动峰度

    Warnings
    --------
    等待polars官方出rolling_kurt

    """
    return x.map_batches(lambda a: roll_kurt(a, d))


def ts_l2_norm(x: Expr, d: int = 5) -> Expr:
    """Euclidean norm

    欧几里得范数"""
    return x.pow(2).rolling_sum(d).sqrt()


def ts_log_diff(x: Expr, d: int = 1) -> Expr:
    """对数差分。log(current value of input or x[t] ) - log(previous value of input or x[t-1]).

    """
    return x.log().diff(d)


def ts_max(x: Expr, d: int = 30) -> Expr:
    """时序滚动最大值"""
    return x.rolling_max(d)


def ts_max_diff(x: Expr, d: int = 30) -> Expr:
    """Returns x - ts_max(x, d)"""
    return x - ts_max(x, d)


def ts_mean(x: Expr, d: int = 5) -> Expr:
    """简单移动平均"""
    return x.rolling_mean(d)


def ts_median(x: Expr, d: int = 5) -> Expr:
    """时序滚动中位数"""
    return x.rolling_median(d)


def ts_min(x: Expr, d: int = 30) -> Expr:
    """时序滚动最小值"""
    return x.rolling_min(d)


def ts_min_diff(x: Expr, d: int = 30) -> Expr:
    """Returns x - ts_min(x, d)"""
    return x - ts_min(x, d)


def ts_min_max_cps(x: Expr, d: int, f: float = 2.0) -> Expr:
    """Returns (ts_min(x, d) + ts_max(x, d)) - f * x"""
    return (ts_min(x, d) + ts_max(x, d)) - f * x


def ts_min_max_diff(x: Expr, d: int, f: float = 0.5) -> Expr:
    """Returns x - f * (ts_min(x, d) + ts_max(x, d))"""
    return x - f * (ts_min(x, d) + ts_max(x, d))


def ts_moment(x: Expr, d: int, k: int = 0) -> Expr:
    """Returns K-th central moment of x for the past d days.

    滚动k阶中心距

    Parameters
    ----------
    x
    d
    k

    """
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), roll_moment, d, k))


def ts_partial_corr(x: Expr, y: Expr, z: Expr, d: int) -> Expr:
    """Returns partial correlation of x, y, z for the past d days.

    滚动偏相关
    """
    return struct([x, y, z]).map_batches(lambda xx: batches_i2_o1([xx.struct[i].to_numpy() for i in range(3)], roll_partial_corr, d))


def ts_percentage(x: Expr, d: int, percentage: float = 0.5) -> Expr:
    """Returns percentile value of x for the past d days.

    滚动百分位数

    Parameters
    ----------
    x
    d
    percentage

    """
    return x.rolling_quantile(percentage, window_size=d)


def ts_product(x: Expr, d: int = 5) -> Expr:
    """时序滚动乘"""
    return x.map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), roll_prod, d))


def ts_rank(x: Expr, d: int = 5) -> Expr:
    """时序滚动排名

    Warnings
    --------
    等待polars官方出rolling_rank

    """
    return x.map_batches(lambda a: roll_rank(a, d, True))


def ts_returns(x: Expr, d: int = 1) -> Expr:
    """简单收益率"""
    return x.pct_change(d)


def ts_scale(x: Expr, d: int = 5) -> Expr:
    """Returns (x – ts_min(x, d)) / (ts_max(x, d) – ts_min(x, d)) + constant

    时序滚动缩放
    """
    a = ts_min(x, d)
    b = ts_max(x, d)
    # return (x - a) / (b - a + TA_EPSILON)
    return when(a != b).then((x - a) / (b - a)).otherwise(0)


def ts_skewness(x: Expr, d: int = 5, bias: bool = False) -> Expr:
    """Return skewness of x for the past d days

    时序滚动偏度

    Parameters
    ----------
    x
    d
    bias
        有偏

    Notes
    -----
    `bias=False`时与`pandas`结果一样

    """
    return x.rolling_skew(d, bias=bias)


def ts_std_dev(x: Expr, d: int = 5, ddof: int = 0) -> Expr:
    """时序滚动标准差

    Parameters
    ----------
    x
    d
    ddof
        自由度

    """
    return x.rolling_std(d, ddof=ddof)


def ts_sum(x: Expr, d: int = 30) -> Expr:
    """时序滚动求和"""
    return x.rolling_sum(d)


def ts_sum_split_by(x: Expr, by: Expr, d: int = 30, k: int = 10) -> Expr:
    """切割论求和。在d窗口范围内以by为依据进行从小到大排序。取最大的N个和最小的N个对应位置的x的和

    Parameters
    ----------
    x
    by
    d
        窗口大小
    k
        最大最小的k个

    Returns
    -------
    Expr
        * top_k
        * bottom_k

    Examples
    --------

    ```python
    df = pl.DataFrame({
        'a': [10, 20, 30, 40, 50, 60, 70, 80, 90, 100],
        'b': [10, 9, 8, 7, 6, 5, 4, 3, 2, 1],
    }).with_columns(
        A=ts_sum_split_by(pl.col('a'), pl.col('b'), 8, 3)
    )

    shape: (10, 3)
    ┌─────┬─────┬───────────────┐
    │ a   ┆ b   ┆ A             │
    │ --- ┆ --- ┆ ---           │
    │ i64 ┆ i64 ┆ struct[2]     │
    ╞═════╪═════╪═══════════════╡
    │ 10  ┆ 10  ┆ {null,null}   │
    │ 20  ┆ 9   ┆ {null,null}   │
    │ 30  ┆ 8   ┆ {null,null}   │
    │ 40  ┆ 7   ┆ {null,null}   │
    │ 50  ┆ 6   ┆ {null,null}   │
    │ 60  ┆ 5   ┆ {null,null}   │
    │ 70  ┆ 4   ┆ {null,null}   │
    │ 80  ┆ 3   ┆ {210.0,60.0}  │
    │ 90  ┆ 2   ┆ {240.0,90.0}  │
    │ 100 ┆ 1   ┆ {270.0,120.0} │
    └─────┴─────┴───────────────┘
    ```

    """
    return struct([x, by]).map_batches(lambda xx: batches_i2_o2([xx.struct[i].to_numpy() for i in range(2)], _sum_split_by, d, k))


def ts_triple_corr(x: Expr, y: Expr, z: Expr, d: int) -> Expr:
    """时序滚动三重相关系数 Returns triple correlation of x, y, z for the past d days.


    """
    return struct([x, y, z]).map_batches(lambda xx: batches_i2_o1([xx.struct[i].to_numpy() for i in range(3)], roll_triple_corr, d))


def ts_weighted_decay(x: Expr, k: float = 0.5) -> Expr:
    """加权衰减 Instead of replacing today’s value with yesterday’s as in ts_delay(x, 1),
    it assigns weighted average of today’s and yesterday’s values with weight on today’s value being k and yesterday’s being (1-k).

    Parameters
    ----------
    x
    k
        衰减系数

    """
    return x.rolling_sum(2, weights=[1 - k, k])


def ts_zscore(x: Expr, d: int = 5) -> Expr:
    """时序滚动zscore"""
    return (x - ts_mean(x, d)) / ts_std_dev(x, d, 0)


def ts_cum_prod_by(r: Expr, v: Expr) -> Expr:
    """带设置的累乘

    可用于市值累乘日收益率得到新市值的需求

    Parameters
    ----------
    r
        收益率。不能出现`null`, `null`需要提前用`1`代替

        `CLOSE/ts_delay(CLOSE, 1)`
    v
        市值。非空时分配指定市值资产

        * 如果非`null`，直接返回`v`
        * 如果`null`，返回`V[-1]*r`

    Returns
    -------
    Expr
        V。累乘后成新的市值

    Examples
    --------

    ```python
    df = pl.DataFrame({
        'r': [1, 2, 3, 4, 5, 6],
        'v': [None, None, 6, None, None, 12],

    }).with_columns(
        V=ts_cum_prod_by(pl.col('r'), pl.col('v'))
    )

    shape: (6, 3)
    ┌─────┬──────┬───────┐
    │ r   ┆ v    ┆ V     │
    │ --- ┆ ---  ┆ ---   │
    │ i64 ┆ i64  ┆ f64   │
    ╞═════╪══════╪═══════╡
    │ 1   ┆ null ┆ null  │
    │ 2   ┆ null ┆ null  │
    │ 3   ┆ 6    ┆ 6.0   │
    │ 4   ┆ null ┆ 24.0  │
    │ 5   ┆ null ┆ 120.0 │
    │ 6   ┆ 12   ┆ 12.0  │
    └─────┴──────┴───────┘
    ```


    """
    return struct([r, v]).map_batches(lambda xx: batches_i2_o1([xx.struct[i].to_numpy().astype(float) for i in range(2)], _cum_prod_by))


def ts_cum_sum_by(r: Expr, v: Expr) -> Expr:
    """带设置的累加

    可用于市值累加日收益得到新市值的需求

    Parameters
    ----------
    r
        收益。不能出现`null`, `null`需要提前用`0`代替

        `CLOSE-ts_delay(CLOSE, 1)`
    v
        市值。非空时分配指定市值资产

        * 如果非`null`，直接返回`v`
        * 如果`null`，返回`V[-1]+r`

    Examples
    --------

    ```python
    df = pl.DataFrame({
        'r': [1, 2, 3, 4, 5, 6],
        'v': [None, None, 6, None, None, 12],

    }).with_columns(
        V=ts_cum_sum_by(pl.col('r'), pl.col('v'))
    )

    shape: (6, 3)
    ┌─────┬──────┬──────┐
    │ r   ┆ v    ┆ V    │
    │ --- ┆ ---  ┆ ---  │
    │ i64 ┆ i64  ┆ f64  │
    ╞═════╪══════╪══════╡
    │ 1   ┆ null ┆ null │
    │ 2   ┆ null ┆ null │
    │ 3   ┆ 6    ┆ 6.0  │
    │ 4   ┆ null ┆ 10.0 │
    │ 5   ┆ null ┆ 15.0 │
    │ 6   ┆ 12   ┆ 12.0 │
    └─────┴──────┴──────┘

    ```

    """
    return struct([r, v]).map_batches(lambda xx: batches_i2_o1([xx.struct[i].to_numpy().astype(float) for i in range(2)], _cum_sum_by))


def ts_regression_resid(y: Expr, x: Expr, d: int) -> Expr:
    """时序滚动回归取残差"""
    return pls.compute_rolling_least_squares(y, x, mode='residuals', add_intercept=True, rolling_kwargs=RollingKwargs(window_size=d, min_periods=d))


def ts_regression_pred(y: Expr, x: Expr, d: int) -> Expr:
    """时序滚动回归取y的预测值
    """
    return pls.compute_rolling_least_squares(y, x, mode='predictions', add_intercept=True, rolling_kwargs=RollingKwargs(window_size=d, min_periods=d))


def ts_regression_intercept(y: Expr, x: Expr, d: int) -> Expr:
    """时序滚动回归取截距
    """
    return pls.compute_rolling_least_squares(y, x, mode='coefficients', add_intercept=True, rolling_kwargs=RollingKwargs(window_size=d, min_periods=d)).struct[1]


def ts_regression_slope(y: Expr, x: Expr, d: int) -> Expr:
    """时序滚动回归取斜率"""
    return pls.compute_rolling_least_squares(y, x, mode='coefficients', add_intercept=True, rolling_kwargs=RollingKwargs(window_size=d, min_periods=d)).struct[0]


def ts_resid(y: Expr, *more_x: Expr, d: int = 30) -> Expr:
    """多元时序滚动回归取残差

    Parameters
    ----------
    y
    *more_x
        多个x
    d

    """
    return pls.compute_rolling_least_squares(y, *more_x, mode='residuals', rolling_kwargs=RollingKwargs(window_size=d, min_periods=d))


def ts_pred(y: Expr, *more_x: Expr, d: int = 30) -> Expr:
    """多元时序滚动回归预测

    Parameters
    ----------
    y
    *more_x
        多个x
    d

    """
    return pls.compute_rolling_least_squares(y, *more_x, mode='predictions', rolling_kwargs=RollingKwargs(window_size=d, min_periods=d))


def ts_weighted_mean(x: Expr, w: Expr, d: int) -> Expr:
    """时序滚动加权平均"""
    return (x * w).rolling_sum(d) / w.rolling_sum(d)


def ts_weighted_sum(x: Expr, w: Expr, d: int) -> Expr:
    """时序滚动加权求和"""
    return (x * w).rolling_sum(d)


def ts_signals_to_size(long_entry: Expr, long_exit: Expr,
                       short_entry: Expr, short_exit: Expr,
                       accumulate: bool = False,
                       action: bool = False) -> Expr:
    """多空信号转持仓。参考于`vectorbt`

    Parameters
    ----------
    long_entry
        多头入场
    long_exit
        多头出场
    short_entry
        空头入场
    short_exit
        空头出场
    accumulate
        遇到重复信号时是否累计
    action
        返回持仓状态还是下单操作

    """
    return struct([long_entry, long_exit, short_entry, short_exit]).map_batches(
        lambda xx: batches_i2_o1([xx.struct[i].to_numpy().astype(float) for i in range(4)],
                                 _signals_to_size, accumulate, action))



"""
通过`import`直接导入或更名的函数

```python
from polars_ta.wq.arithmetic import add as ADD  # noqa
from polars_ta.wq.arithmetic import divide as DIV  # noqa
from polars_ta.wq.arithmetic import multiply as MUL  # noqa
from polars_ta.wq.arithmetic import subtract as SUB  # noqa
from polars_ta.wq.time_series import ts_max as MAX  # noqa
from polars_ta.wq.time_series import ts_min as MIN  # noqa
from polars_ta.wq.time_series import ts_sum as SUM  # noqa
```

"""
from polars import Expr

from polars_ta.wq.arithmetic import add as ADD  # noqa
from polars_ta.wq.arithmetic import divide as DIV  # noqa
from polars_ta.wq.arithmetic import multiply as MUL  # noqa
from polars_ta.wq.arithmetic import subtract as SUB  # noqa
from polars_ta.wq.time_series import ts_arg_max
from polars_ta.wq.time_series import ts_arg_min
from polars_ta.wq.time_series import ts_max as MAX  # noqa
from polars_ta.wq.time_series import ts_min as MIN  # noqa
from polars_ta.wq.time_series import ts_sum as SUM  # noqa


def MAXINDEX(close: Expr, timeperiod: int = 30) -> Expr:
    """

    Notes
    -----
    Comparing to `ts_arg_max` this also marks the abs. position of the max value

    与ts_arg_max的区别是，标记了每个区间最大值的绝对位置，可用来画图标记


    Examples
    --------
    ```python
    from polars_ta.ta import MAXINDEX as ta_MAXINDEX
    from polars_ta.talib import MAXINDEX as talib_MAXINDEX
    from polars_ta.wq import ts_arg_max

    df = pl.DataFrame({
        'a': [6, 2, 8, 5, 9, 4],
    }).with_columns(
        out1=ts_arg_max(pl.col('a'), 3),
        out2=ta_MAXINDEX(pl.col('a'), 3),
        out3=talib_MAXINDEX(pl.col('a'), 3),
    )
    shape: (6, 4)
    ┌─────┬──────┬──────┬──────┐
    │ a   ┆ out1 ┆ out2 ┆ out3 │
    │ --- ┆ ---  ┆ ---  ┆ ---  │
    │ i64 ┆ u16  ┆ i64  ┆ i32  │
    ╞═════╪══════╪══════╪══════╡
    │ 6   ┆ null ┆ null ┆ 0    │
    │ 2   ┆ null ┆ null ┆ 0    │
    │ 8   ┆ 0    ┆ 2    ┆ 2    │
    │ 5   ┆ 1    ┆ 2    ┆ 2    │
    │ 9   ┆ 0    ┆ 4    ┆ 4    │
    │ 4   ┆ 1    ┆ 4    ┆ 4    │
    └─────┴──────┴──────┴──────┘
    ```
    """
    a = close.cum_count()
    b = ts_arg_max(close, timeperiod)
    return a - b - 1


def MININDEX(close: Expr, timeperiod: int = 30) -> Expr:
    """



    Examples
    --------
    ```python
    from polars_ta.ta import MININDEX as ta_MININDEX
    from polars_ta.talib import MININDEX as talib_MININDEX
    from polars_ta.wq import ts_arg_min

    df = pl.DataFrame({
        'a': [6, 2, 8, 5, 9, 4],

    }).with_columns(
        out1=ts_arg_min(pl.col('a'), 3),
        out2=ta_MININDEX(pl.col('a'), 3),
        out3=talib_MININDEX(pl.col('a'), 3),
    )
    shape: (6, 4)
    ┌─────┬──────┬──────┬──────┐
    │ a   ┆ out1 ┆ out2 ┆ out3 │
    │ --- ┆ ---  ┆ ---  ┆ ---  │
    │ i64 ┆ u16  ┆ i64  ┆ i32  │
    ╞═════╪══════╪══════╪══════╡
    │ 6   ┆ null ┆ null ┆ 0    │
    │ 2   ┆ null ┆ null ┆ 0    │
    │ 8   ┆ 1    ┆ 1    ┆ 1    │
    │ 5   ┆ 2    ┆ 1    ┆ 1    │
    │ 9   ┆ 1    ┆ 3    ┆ 3    │
    │ 4   ┆ 0    ┆ 5    ┆ 5    │
    └─────┴──────┴──────┴──────┘
    ```
    """
    a = close.cum_count()
    b = ts_arg_min(close, timeperiod)
    return a - b - 1



"""
与`WorldQuant Alpha101`的区别是添加了`cs_`前缀

由于截面与时序的使用方式不同，在自动化工具中如果不在名字上做区分就得手工注册，反而要麻烦些

"""

import polars_ols as pls
from polars import Expr, when
from polars_ols import OLSKwargs

# In the original version, the function names are not prefixed with `cs_`,
# here we add it to prevent confusion
# 原版函数名都没有加`cs_`, 这里统一加一防止混淆


_ols_kwargs = OLSKwargs(null_policy='drop', solve_method='svd')


def cs_one_side(x: Expr, is_long: bool = True) -> Expr:
    """Shifts all instruments up or down so that the Alpha becomes long-only or short-only
(if side = short), respectively.

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, -15, -7, 0, 20],
        'b': [None, 15, 7, 0, 20],
    }).with_columns(
        out1=cs_one_side(pl.col('a'), True),
        out2=cs_one_side(pl.col('a'), False),
        out3=cs_one_side(pl.col('b'), True),
        out4=cs_one_side(pl.col('b'), False),
    )
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────┬──────┐
    │ a    ┆ b    ┆ out1 ┆ out2 ┆ out3 ┆ out4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ i64  ┆ i64  ┆ i64  ┆ i64  │
    ╞══════╪══════╪══════╪══════╪══════╪══════╡
    │ null ┆ null ┆ null ┆ null ┆ null ┆ null │
    │ -15  ┆ 15   ┆ 0    ┆ -35  ┆ 15   ┆ -5   │
    │ -7   ┆ 7    ┆ 8    ┆ -27  ┆ 7    ┆ -13  │
    │ 0    ┆ 0    ┆ 15   ┆ -20  ┆ 0    ┆ -20  │
    │ 20   ┆ 20   ┆ 35   ┆ 0    ┆ 20   ┆ 0    │
    └──────┴──────┴──────┴──────┴──────┴──────┘
    ```

    """
    if is_long:
        return when(x.min() < 0).then(x - x.min()).otherwise(x)
    else:
        return when(x.max() > 0).then(x - x.max()).otherwise(x)


def cs_rank(x: Expr, pct: bool = True) -> Expr:
    """排名。Ranks the input among all the instruments and returns an equally distributed number between 0.0 and 1.0. For precise sort, use the rate as 0.

    Parameters
    ----------
    x
    pct
        * True: 排名百分比。范围：[0,1]
        * False: 排名。范围：[1,+inf)

    Examples
    --------

    ```python
    df = pl.DataFrame({
        'a': [None, 1, 1, 1, 2, 2, 3, 10],
    }).with_columns(
        out1=cs_rank(pl.col('a'), True),
        out2=cs_rank(pl.col('a'), False),
    )

    shape: (8, 3)
    ┌──────┬──────────┬──────┐
    │ a    ┆ out1     ┆ out2 │
    │ ---  ┆ ---      ┆ ---  │
    │ i64  ┆ f64      ┆ u32  │
    ╞══════╪══════════╪══════╡
    │ null ┆ null     ┆ null │
    │ 1    ┆ 0.0      ┆ 1    │
    │ 1    ┆ 0.0      ┆ 1    │
    │ 1    ┆ 0.0      ┆ 1    │
    │ 2    ┆ 0.5      ┆ 4    │
    │ 2    ┆ 0.5      ┆ 4    │
    │ 3    ┆ 0.833333 ┆ 6    │
    │ 10   ┆ 1.0      ┆ 7    │
    └──────┴──────────┴──────┘
    ```

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#rankx-rate2

    """
    if pct:
        # (x-x.min)/(x.max-x.min)
        r = x.rank(method='min') - 1
        return r / r.max()
    else:
        return x.rank(method='min')


def cs_scale(x: Expr, scale_: float = 1, long_scale: float = 1, short_scale: float = 1) -> Expr:
    """Scales input to booksize. We can also scale the long positions and short positions to separate scales by mentioning additional parameters to the operator.

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, -15, -7, 0, 20],
    }).with_columns(
        out1=cs_scale(pl.col('a'), 1),
        out2=cs_scale(pl.col('a'), 1, 2, 3),
    )
    shape: (5, 3)
    ┌──────┬───────────┬───────────┐
    │ a    ┆ out1      ┆ out2      │
    │ ---  ┆ ---       ┆ ---       │
    │ i64  ┆ f64       ┆ f64       │
    ╞══════╪═══════════╪═══════════╡
    │ null ┆ null      ┆ null      │
    │ -15  ┆ -0.357143 ┆ -2.045455 │
    │ -7   ┆ -0.166667 ┆ -0.954545 │
    │ 0    ┆ 0.0       ┆ 0.0       │
    │ 20   ┆ 0.47619   ┆ 2.0       │
    └──────┴───────────┴───────────┘
    ```

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#scale-x-scale1-longscale1-shortscale1

    """
    if long_scale != 1 or short_scale != 1:
        L = x.clip(lower_bound=0)  # 全正数
        S = x.clip(upper_bound=0)  # 全负数
        L = when(L.sum() == 0).then(0).otherwise(L / L.sum())
        S = when(S.sum() == 0).then(0).otherwise(S / S.sum())  # 负数/负数=正数
        return L * long_scale - S * short_scale
    else:
        return x / x.abs().sum() * scale_


def cs_scale_down(x: Expr, constant: int = 0) -> Expr:
    """Scales all values in each day proportionately between 0 and 1 such that minimum value maps to 0 and maximum value maps to 1.
    constant is the offset by which final result is subtracted

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 15, 7, 0, 20],
    }).with_columns(
        out1=cs_scale_down(pl.col('a'), 0),
        out2=cs_scale_down(pl.col('a'), 1),
    )
    shape: (5, 3)
    ┌──────┬──────┬───────┐
    │ a    ┆ out1 ┆ out2  │
    │ ---  ┆ ---  ┆ ---   │
    │ i64  ┆ f64  ┆ f64   │
    ╞══════╪══════╪═══════╡
    │ null ┆ null ┆ null  │
    │ 15   ┆ 0.75 ┆ -0.25 │
    │ 7    ┆ 0.35 ┆ -0.65 │
    │ 0    ┆ 0.0  ┆ -1.0  │
    │ 20   ┆ 1.0  ┆ 0.0   │
    └──────┴──────┴───────┘
    ```

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#scale_downxconstant0

    """
    return (x - x.min()) / (x.max() - x.min()) - constant


def cs_truncate(x: Expr, max_percent: float = 0.01) -> Expr:
    """Operator truncates all values of x to maxPercent. Here, maxPercent is in decimal notation

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [3, 7, 20, 6],
    }).with_columns(
        out=cs_truncate(pl.col('a'), 0.5),
    )
    shape: (4, 2)
    ┌─────┬─────┐
    │ a   ┆ out │
    │ --- ┆ --- │
    │ i64 ┆ i64 │
    ╞═════╪═════╡
    │ 3   ┆ 3   │
    │ 7   ┆ 7   │
    │ 20  ┆ 18  │
    │ 6   ┆ 6   │
    └─────┴─────┘
    ```

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#truncatexmaxpercent001

    """
    return x.clip(upper_bound=x.sum() * max_percent)


def cs_fill_except_all_null(x: Expr, value=0) -> Expr:
    """全为`null`时，保持`null`，反之`null`填充为`value`

    Examples
    --------

    ```python
    df = pl.DataFrame({
        'a': [1, 2, None, 4, None],
        'b': [None, None, None, None, None],
    }).with_columns(
        A=cs_fill_except_all_null(pl.col('a')),
        B=cs_fill_except_all_null(pl.col('b')),
    )

    shape: (5, 4)
    ┌──────┬──────┬─────┬──────┐
    │ a    ┆ b    ┆ A   ┆ B    │
    │ ---  ┆ ---  ┆ --- ┆ ---  │
    │ i64  ┆ null ┆ i64 ┆ i32  │
    ╞══════╪══════╪═════╪══════╡
    │ 1    ┆ null ┆ 1   ┆ null │
    │ 2    ┆ null ┆ 2   ┆ null │
    │ null ┆ null ┆ 0   ┆ null │
    │ 4    ┆ null ┆ 4   ┆ null │
    │ null ┆ null ┆ 0   ┆ null │
    └──────┴──────┴─────┴──────┘
    ```

    Notes
    -----
    在权重矩阵中使用时。一定要保证所有股票都在，停牌不能被过滤了

    """
    return when(x.is_not_null().sum() == 0).then(x).otherwise(x.fill_null(value))


def cs_regression_neut(y: Expr, x: Expr) -> Expr:
    """一元回归残差"""
    return pls.compute_least_squares(y, x, add_intercept=True, mode='residuals', ols_kwargs=_ols_kwargs)


def cs_regression_proj(y: Expr, x: Expr) -> Expr:
    """一元回归预测"""
    return pls.compute_least_squares(y, x, add_intercept=True, mode='predictions', ols_kwargs=_ols_kwargs)


def cs_qcut(x: Expr, q: int = 10) -> Expr:
    """等频分箱 Convert float values into indexes for user-specified buckets. Bucket is useful for creating group values, which can be passed to group operators as input.

    Parameters
    ----------
    x
    q
        按频率分成`q`份

    Examples
    --------

    ```python
    df = pl.DataFrame({
        'a': [None, 1, 1, 1, 2, 2, 3, 10],
    }).with_columns(
        out1=cs_qcut(pl.col('a'), 10),
        out2=pl.col('a').map_batches(lambda x: pd.qcut(x, 10, labels=False, duplicates='drop')),
    )
    shape: (8, 3)
    ┌──────┬──────┬──────┐
    │ a    ┆ out1 ┆ out2 │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ u32  ┆ f64  │
    ╞══════╪══════╪══════╡
    │ null ┆ null ┆ NaN  │
    │ 1    ┆ 0    ┆ 0.0  │
    │ 1    ┆ 0    ┆ 0.0  │
    │ 1    ┆ 0    ┆ 0.0  │
    │ 2    ┆ 3    ┆ 1.0  │
    │ 2    ┆ 3    ┆ 1.0  │
    │ 3    ┆ 7    ┆ 4.0  │
    │ 10   ┆ 8    ┆ 5.0  │
    └──────┴──────┴──────┘

    ```

    Warnings
    --------
    目前与`pd.qcut`结果不同，等官方改进

    """
    return x.qcut(q, allow_duplicates=True).to_physical()



from polars import Expr, when, Boolean, Int32


def cut(x: Expr, b: float, *more_bins) -> Expr:
    """分箱

    Parameters
    ----------
    x
    b
    *more_bins

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 1, 1, 1, 2, 2, 3, 10],
    }).with_columns(
        out1=cut(pl.col('a'), 2, 5, 20),
    )
    shape: (8, 2)
    ┌──────┬──────┐
    │ a    ┆ out1 │
    │ ---  ┆ ---  │
    │ i64  ┆ u32  │
    ╞══════╪══════╡
    │ null ┆ null │
    │ 1    ┆ 0    │
    │ 1    ┆ 0    │
    │ 1    ┆ 0    │
    │ 2    ┆ 0    │
    │ 2    ┆ 0    │
    │ 3    ┆ 1    │
    │ 10   ┆ 2    │
    └──────┴──────┘

    ```

    """
    return x.cut([b, *more_bins]).to_physical()


def clamp(x: Expr, lower: float = 0, upper: float = 0, inverse: bool = False, mask: float = None) -> Expr:
    """Limits input value between lower and upper bound in inverse = false mode (which is default). Alternatively, when inverse = true, values between bounds are replaced with mask, while values outside bounds are left as is.

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 1, 2, 3, 4, 5, 6],
    }).with_columns(
        out1=clamp(pl.col('a'), 2, 5, False),
        out2=clamp(pl.col('a'), 2, 5, True),
    )
    shape: (7, 3)
    ┌──────┬──────┬──────┐
    │ a    ┆ out1 ┆ out2 │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ i64  │
    ╞══════╪══════╪══════╡
    │ null ┆ null ┆ null │
    │ 1    ┆ 2    ┆ 1    │
    │ 2    ┆ 2    ┆ null │
    │ 3    ┆ 3    ┆ null │
    │ 4    ┆ 4    ┆ null │
    │ 5    ┆ 5    ┆ null │
    │ 6    ┆ 5    ┆ 6    │
    └──────┴──────┴──────┘

    ```

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#clampx-lower-0-upper-0-inverse-false-mask

    """
    if inverse:
        cond = (x >= lower) & (x <= upper)
        return when(~cond).then(x).otherwise(mask)
    else:
        return x.clip(lower, upper)


# def filter_(x: Expr, h: str = "1, 2, 3, 4", t: str = "0.5") -> Expr:
#     """Used to filter the value and allows to create filters like linear or exponential decay."""
#     raise


def keep(x: Expr, f: float, period: int = 5) -> Expr:
    """This operator outputs value x when f changes and continues to do that for “period” days after f stopped changing. After “period” days since last change of f, NaN is output."""
    raise


def left_tail(x: Expr, maximum: float = 0) -> Expr:
    """NaN everything greater than maximum, maximum should be constant.

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 1, 2, 3, 4, 5],
    }).with_columns(
        out=left_tail(pl.col('a'), 3),
    )
    shape: (6, 2)
    ┌──────┬──────┐
    │ a    ┆ out  │
    │ ---  ┆ ---  │
    │ i64  ┆ i64  │
    ╞══════╪══════╡
    │ null ┆ null │
    │ 1    ┆ 1    │
    │ 2    ┆ 2    │
    │ 3    ┆ 3    │
    │ 4    ┆ null │
    │ 5    ┆ null │
    └──────┴──────┘
    ```

    See Also
    --------
    tail

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#left_tail

    """
    return when(x <= maximum).then(x).otherwise(None)


# def pasteurize(x: Expr) -> Expr:
#     """Set to NaN if x is INF or if the underlying instrument is not in the Alpha universe"""
#     # TODO: 不在票池中的的功能无法表示
#     # TODO: 与purify好像没啥区别
#     return when(x.is_finite()).then(x).otherwise(None)


def purify(x: Expr) -> Expr:
    """Clear infinities (+inf, -inf) by replacing with null.

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 1., 2., float('nan'), float('inf'), float('-inf')],
    }).with_columns(
        out=purify(pl.col('a')),
    )
    shape: (6, 2)
    ┌──────┬──────┐
    │ a    ┆ out  │
    │ ---  ┆ ---  │
    │ f64  ┆ f64  │
    ╞══════╪══════╡
    │ null ┆ null │
    │ 1.0  ┆ 1.0  │
    │ 2.0  ┆ 2.0  │
    │ NaN  ┆ null │
    │ inf  ┆ null │
    │ -inf ┆ null │
    └──────┴──────┘
    ```

    """
    return when(x.is_finite()).then(x).otherwise(None)


def fill_nan(x: Expr) -> Expr:
    """填充`nan`为`null`"""
    return x.fill_nan(None)


def fill_null(x: Expr, value=0) -> Expr:
    """填充`null`为`value`"""
    return x.fill_null(value)


def right_tail(x: Expr, minimum: float = 0) -> Expr:
    """NaN everything less than minimum, minimum should be constant.

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 1, 2, 3, 4, 5],
    }).with_columns(
        out=right_tail(pl.col('a'), 3),
    )
    shape: (6, 2)
    ┌──────┬──────┐
    │ a    ┆ out  │
    │ ---  ┆ ---  │
    │ i64  ┆ i64  │
    ╞══════╪══════╡
    │ null ┆ null │
    │ 1    ┆ null │
    │ 2    ┆ null │
    │ 3    ┆ 3    │
    │ 4    ┆ 4    │
    │ 5    ┆ 5    │
    └──────┴──────┘
    ```

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#right_tail

    """
    return when(x >= minimum).then(x).otherwise(None)


def sigmoid(x: Expr) -> Expr:
    """Returns 1 / (1 + exp(-x))"""
    return 1 / (1 + (-x).exp())


def tail(x: Expr, lower: float = 0, upper: float = 0, newval: float = 0) -> Expr:
    """If (x > lower AND x < upper) return newval, else return x. Lower, upper, newval should be constants.

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 1, 2, 3, 4, 5, 6],
    }).with_columns(
        out=tail(pl.col('a'), 2, 5, 1),
    )
    shape: (7, 2)
    ┌──────┬──────┐
    │ a    ┆ out  │
    │ ---  ┆ ---  │
    │ i64  ┆ i64  │
    ╞══════╪══════╡
    │ null ┆ null │
    │ 1    ┆ 1    │
    │ 2    ┆ 2    │
    │ 3    ┆ 1    │
    │ 4    ┆ 1    │
    │ 5    ┆ 5    │
    │ 6    ┆ 6    │
    └──────┴──────┘

    ```

    See Also
    --------
    clamp

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#tail

    """

    cond = (x > lower) & (x < upper)
    return when(~cond | x.is_null()).then(x).otherwise(newval)


def int_(a: Expr) -> Expr:
    """convert bool to int"""
    return a.cast(Int32)


def bool_(a: Expr) -> Expr:
    """convert int to bool"""
    return a.cast(Boolean)



import numpy as np
from polars import Expr, Series, fold, any_horizontal
from polars import max_horizontal, sum_horizontal, min_horizontal, mean_horizontal


def abs_(x: Expr) -> Expr:
    """绝对值"""
    if isinstance(x, (Expr, Series)):
        return x.abs()
    else:
        return np.abs(x)


def add(a: Expr, b: Expr, *args) -> Expr:
    """水平多列加

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 2, 3, 4, None],
        'b': [5, None, 3, 2, None],
        'c': [1, 1, None, 1, None],
    }).with_columns(
        out=add(pl.col('a'), pl.col('b'), pl.col('c'))
    )
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ a    ┆ b    ┆ c    ┆ out  │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ i64  ┆ i64  │
    ╞══════╪══════╪══════╪══════╡
    │ null ┆ 5    ┆ 1    ┆ 6    │
    │ 2    ┆ null ┆ 1    ┆ 3    │
    │ 3    ┆ 3    ┆ null ┆ 6    │
    │ 4    ┆ 2    ┆ 1    ┆ 7    │
    │ null ┆ null ┆ null ┆ null │
    └──────┴──────┴──────┴──────┘

    ```

    Notes
    -----
    全`null`时返回`null`

    """
    # # 全null时返回0
    # return sum_horizontal(a, b, *args)

    _args = [a, b] + list(args)
    return fold(acc=any_horizontal(_args) - 1, function=lambda acc, x: acc + x.fill_null(0), exprs=_args)


def arc_cos(x: Expr) -> Expr:
    """反余弦"""
    return x.arccos()


def arc_sin(x: Expr) -> Expr:
    """反正弦"""
    return x.arcsin()


def arc_tan(x: Expr) -> Expr:
    """反正切"""
    return x.arctan()


def ceiling(x: Expr) -> Expr:
    """向上取整"""
    return x.ceil()


def cos(x: Expr) -> Expr:
    """余弦"""
    return x.cos()


def cosh(x: Expr) -> Expr:
    """双曲余弦"""
    return x.cosh()


def densify(x: Expr) -> Expr:
    raise


def divide(x: Expr, y: Expr) -> Expr:
    """x/y"""
    return x / y


def exp(x: Expr) -> Expr:
    """自然指数函数"""
    return x.exp()


def expm1(x: Expr) -> Expr:
    """对数收益率 转 简单收益率 convert log return to simple return"""
    return x.exp() - 1


def floor(x: Expr) -> Expr:
    """向下取整"""
    return x.floor()


def fraction(x: Expr) -> Expr:
    """小数部分
    This operator removes the whole number part and returns the remaining fraction part with sign.

    Examples
    --------

    ```python
    df = pl.DataFrame({
        'a': [-2.5, -1.2, -1., None, 2., 3.2],
    }).with_columns(
        out=fraction(pl.col('a'))
    )

    shape: (6, 2)
    ┌──────┬──────┐
    │ a    ┆ out  │
    │ ---  ┆ ---  │
    │ f64  ┆ f64  │
    ╞══════╪══════╡
    │ -2.5 ┆ -0.5 │
    │ -1.2 ┆ -0.2 │
    │ -1.0 ┆ -0.0 │
    │ null ┆ null │
    │ 2.0  ┆ 0.0  │
    │ 3.2  ┆ 0.2  │
    └──────┴──────┘
    ```

    Notes
    -----
    按小学时的定义，负数`-1.2`的整数部分是`-2`,小数部分是`0.8`，而这有所不同

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#fractionx

    """
    return x.sign() * (x.abs() % 1)


def inverse(x: Expr) -> Expr:
    """1/x"""
    return 1 / x


def log(x: Expr) -> Expr:
    """e为底的对数"""
    if isinstance(x, (Expr, Series)):
        return x.log()
    else:
        return np.log(x)


def log10(x: Expr) -> Expr:
    """10为底的对数"""
    return x.log10()


def log1p(x: Expr) -> Expr:
    """简单收益率 转 对数收益率 convert simple return to log return

    log(x+1)
    """
    return x.log1p()


def max_(a: Expr, b: Expr, *args) -> Expr:
    """水平多列最大值 Maximum value of all inputs. At least 2 inputs are required."""
    return max_horizontal(a, b, *args)


def mean(a: Expr, b: Expr, *args) -> Expr:
    """水平多列均值"""
    return mean_horizontal(a, b, *args)


def min_(a: Expr, b: Expr, *args) -> Expr:
    """水平多列最小值 Maximum value of all inputs. At least 2 inputs are required."""
    return min_horizontal(a, b, *args)


def mod(x: Expr, y: Expr) -> Expr:
    """x%y"""
    return x % y


def multiply(a: Expr, b: Expr, *args) -> Expr:
    """水平多列乘 Multiply all inputs. At least 2 inputs are required.

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 2, 3, 4, None],
        'b': [5, None, 3, 2, None],
        'c': [1, 1, None, 1, None],
    }).with_columns(
        out=multiply(pl.col('a'), pl.col('b'), pl.col('c'))
    )

    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ a    ┆ b    ┆ c    ┆ out  │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ i64  ┆ i64  │
    ╞══════╪══════╪══════╪══════╡
    │ null ┆ 5    ┆ 1    ┆ 5    │
    │ 2    ┆ null ┆ 1    ┆ 2    │
    │ 3    ┆ 3    ┆ null ┆ 9    │
    │ 4    ┆ 2    ┆ 1    ┆ 8    │
    │ null ┆ null ┆ null ┆ null │
    └──────┴──────┴──────┴──────┘

    ```

    Notes
    -----
    全`null`时返回`null`

    """
    _args = [a, b] + list(args)

    # # 全null返回1
    # return fold(acc=1, function=lambda acc, x: acc * x.fill_null(1), exprs=_args)
    return fold(acc=any_horizontal(_args), function=lambda acc, x: acc * x.fill_null(1), exprs=_args)


def power(x: Expr, y: Expr) -> Expr:
    """x ** y"""
    return x.pow(y)


def reverse(x: Expr) -> Expr:
    """-x"""
    return -x


def round_(x: Expr, decimals: int = 0) -> Expr:
    """四舍五入 Round input to closest integer.

    Parameters
    ----------
    x
    decimals
        Number of decimals to round to.

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 3.5, 4.5, -3.5, -4.5],
    }).with_columns(
        out1=round_(pl.col('a'), 0),
        out2=pl.col('a').map_elements(lambda x: round(x, 0), return_dtype=pl.Float64),
    )
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ a    ┆ out1 ┆ out2 │
    │ ---  ┆ ---  ┆ ---  │
    │ f64  ┆ f64  ┆ f64  │
    ╞══════╪══════╪══════╡
    │ null ┆ null ┆ null │
    │ 3.5  ┆ 4.0  ┆ 4.0  │
    │ 4.5  ┆ 5.0  ┆ 4.0  │
    │ -3.5 ┆ -4.0 ┆ -4.0 │
    │ -4.5 ┆ -5.0 ┆ -4.0 │
    └──────┴──────┴──────┘
    ```

    Notes
    -----
    四舍五入，不是四舍六入五取偶（银行家舍入）

    """
    return x.round(decimals)


def round_down(x: Expr, f: int = 1) -> Expr:
    """小于输入的f的最大倍数 Round input to greatest multiple of f less than input

    Parameters
    ----------
    x
    f

    Examples
    --------

    ```python
    df = pl.DataFrame({
        'a': [None, 3.5, 4.5, -3.5, -4.5],
    }).with_columns(
        out=round_down(pl.col('a'), 2),
    )
    shape: (5, 2)
    ┌──────┬──────┐
    │ a    ┆ out  │
    │ ---  ┆ ---  │
    │ f64  ┆ f64  │
    ╞══════╪══════╡
    │ null ┆ null │
    │ 3.5  ┆ 2.0  │
    │ 4.5  ┆ 4.0  │
    │ -3.5 ┆ -4.0 │
    │ -4.5 ┆ -6.0 │
    └──────┴──────┘
    ```

    """
    if f == 1:
        return x // 1
    else:
        return x // f * f


def s_log_1p(x: Expr) -> Expr:
    """sign(x) * log10(1 + abs(x))

    Examples
    --------
    ```python
    df = pl.DataFrame({
        'a': [None, 9, -9, 99, -99],
    }).with_columns(
        out1=s_log_1p(pl.col('a')),
    )
    shape: (5, 2)
    ┌──────┬──────┐
    │ a    ┆ out1 │
    │ ---  ┆ ---  │
    │ i64  ┆ f64  │
    ╞══════╪══════╡
    │ null ┆ null │
    │ 9    ┆ 1.0  │
    │ -9   ┆ -1.0 │
    │ 99   ┆ 2.0  │
    │ -99  ┆ -2.0 │
    └──────┴──────┘

    ```

    Notes
    -----
    从`wq`示例可以看出，log的底数是10，而不是e

    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#s_log_1px

    """
    return (x.abs() + 1).log10() * x.sign()


def sign(x: Expr) -> Expr:
    """符号"""
    if isinstance(x, (Expr, Series)):
        return x.sign()
    else:
        return np.sign(x)


def signed_power(x: Expr, y: Expr) -> Expr:
    """x raised to the power of y such that final result preserves sign of x.


    References
    ----------
    https://platform.worldquantbrain.com/learn/operators/detailed-operator-descriptions#signed_powerx-y

    """
    if isinstance(y, (int, float)):
        if y == 1:
            return x.abs() * x.sign()
        elif y == 0:
            return x.sign()

    return x.abs().pow(y) * x.sign()


def sin(x: Expr) -> Expr:
    """正弦"""
    return x.sin()


def sinh(x: Expr) -> Expr:
    """双曲正弦"""
    return x.sinh()


def softsign(x: Expr) -> Expr:
    """softsign是 tanh激活函数的另一个替代选择"""
    return x / (1 + x.abs())


def sqrt(x: Expr) -> Expr:
    """平方根"""
    return x.sqrt()


def subtract(x: Expr, y: Expr) -> Expr:
    """x-y"""
    return x - y


def tan(x: Expr) -> Expr:
    """正切"""
    return x.tan()


def tanh(x: Expr) -> Expr:
    """双曲正切"""
    return x.tanh()


def var(a: Expr, b: Expr, *args) -> Expr:
    """水平多列方差"""
    _args = [a, b] + list(args)
    _mean = mean_horizontal(_args)
    _sum = sum_horizontal([(expr - _mean) ** 2 for expr in _args])
    return _sum


def std(a: Expr, b: Expr, *args) -> Expr:
    """水平多列标准差"""
    return var(a, b, *args).sqrt()
